// override.cpp - LCTextOut
#include "override.h"

#if defined(__SSE3__)
#include <pmmintrin.h>
#elif defined(__SSE2__)
#include <emmintrin.h>
#endif
#include <malloc.h>		// _alloca
#include <mbctype.h>	// _getmbcp

#pragma comment(lib, "Gdi32.lib")
#pragma comment(lib, "User32.lib")

UINT g_Quality = 0;
UINT g_Weight = 0;
UINT g_UseSubPixel = 0;
UINT g_SubPixelDirection = 0;
float g_MaxHeight = 0;
BOOL g_IsWinXPorLater = FALSE;

int  g_Scale = 1;

#define MAX_EXCLUDES 32
struct StringHash {
	DWORD	hash;
	WCHAR	name[LF_FACESIZE];
	
	StringHash() {}
	StringHash(LPCWSTR s)
	{
		Init(s);
	}
	void Init(LPCWSTR s)
	{
		wcsncpy(name, s, LF_FACESIZE);
		_UpdateHash();
	}
	bool operator ==(const StringHash& x) const
	{
		return !(hash != x.hash || wcscmp(name, x.name) != 0);
	}
	void _UpdateHash()
	{
		DWORD dw = 0;
		LPWSTR p;
		for (p = name; *p; p++) {
			dw <<= 3;
			dw ^= *p;
		}
		hash = dw;
	}
} g_ExcludeList[MAX_EXCLUDES];


// ExtTextOutW̃rbg}bvLbV
struct {
	HDC	hdc;
	HBITMAP	hbmp;
	BYTE*	lpPixels;
	SIZE	dibSize;
	CRITICAL_SECTION  cs;
} g_Cache = { NULL,NULL,NULL, {0,0},{0} };

struct CacheLock {
	CacheLock()  { EnterCriticalSection(&g_Cache.cs); }
	~CacheLock() { LeaveCriticalSection(&g_Cache.cs); }
};

void  CacheInit()
{
	InitializeCriticalSection(&g_Cache.cs);
}

void  CacheTerm()
{
	if(g_Cache.hdc) DeleteDC(g_Cache.hdc);
	if(g_Cache.hbmp) DeleteObject(g_Cache.hbmp);
	g_Cache.hdc = NULL;
	g_Cache.hbmp = NULL;
	g_Cache.lpPixels = NULL;
	DeleteCriticalSection(&g_Cache.cs);
}

void CacheFillSolidRect(COLORREF rgb, const RECT* lprc)
{
	rgb = RGB(GetBValue(rgb), GetGValue(rgb), GetRValue(rgb));
	DWORD* p = (DWORD*)g_Cache.lpPixels;
	//ULONG_PTR pend = (ULONG_PTR)p + (g_Cache.dibSize.cx * g_Cache.dibSize.cy);
	unsigned long* pend = (unsigned long*)p + (g_Cache.dibSize.cx * g_Cache.dibSize.cy);
	// lpPixels͕̒lɂȂ鎖̂unsignedŔr
	//while((ULONG_PTR)p < pend)
	while((unsigned long*)p < pend)
		*p++ = rgb;
}


// INIt@Cǂݍ
void LoadSettings(HMODULE hModule){
	WCHAR szFileName[MAX_PATH + 1] = L"";
	// DLL̃tpX擾
	GetModuleFileName(hModule, szFileName, MAX_PATH);
	// gq".ini"ɒu
	lstrcpy(szFileName + lstrlen(szFileName) - 3, L"ini");
	// eݒǂݍ
	// INIt@C̗:
	// [General]
	// Quality=1
	// Weight=0
	// UseSubPixel=0
	// SubPixelDirection=0
	g_Quality = GetPrivateProfileInt(L"General",L"Quality",2,szFileName);
	g_Weight = GetPrivateProfileInt(L"General",L"Weight",0,szFileName);
	g_UseSubPixel = GetPrivateProfileInt(L"General",L"UseSubPixel",0,szFileName);
	g_SubPixelDirection = GetPrivateProfileInt(L"General",L"SubPixelDirection",0,szFileName);
	g_MaxHeight = (float)GetPrivateProfileInt(L"General",L"MaxHeight",0,szFileName);

	// OS̃o[WXPȍ~ǂ
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx (&osvi);
	g_IsWinXPorLater = ((osvi.dwMajorVersion > 5) || ((osvi.dwMajorVersion == 5) && (osvi.dwMinorVersion >= 1)));

	// [Exclude]ZNV珜OtHgXgǂݍ
	memset(&g_ExcludeList, 0, sizeof(g_ExcludeList));

	WCHAR  buffer [4096];
	WCHAR* p = buffer;
	UINT   nb = 0;
	*p = '\0';
	GetPrivateProfileSection(L"Exclude", p, 4096, szFileName);
	while(*p) {
		bool b = false;
		StringHash* exclist = g_ExcludeList;
		StringHash str(p);
		for(UINT i=0 ; i < nb ; i++) {
			if(exclist[i] == str) {
				b = true;
				break;
			}
		}
		if(!b) {
			if(nb < MAX_EXCLUDES) {
				exclist[nb++] = str;
			}
		}
		while(*p) p++;
		p++;
	}

	switch(g_Quality) {
	case 3:  g_Scale = 4; break;
	case 2:  g_Scale = 3; break;
	case 1:  g_Scale = 2; break;
	default: g_Scale = 1; break;
	}
}

BOOL IsFontExcluded(const LPCWSTR lpszFaceName){
	StringHash* exclist = g_ExcludeList;
	StringHash str(lpszFaceName);
	for(UINT i=0 ; i < MAX_EXCLUDES && exclist[i].hash ; i++) {
		if(exclist[i] == str)
			return TRUE;
	}
	return FALSE;
}

#define HOOK_DEFINE(dll, rettype, name, argtype, vars) \
	rettype WINAPI IMPL_##name argtype \
	{ \
	if (g_Quality == 0) return ORIG_##name vars; \
	int grmode = GetGraphicsMode(hdc); \
	if(grmode != GM_ADVANCED) \
		SetGraphicsMode(hdc, GM_ADVANCED); \
	XFORM xformOrig; \
	GetWorldTransform(hdc, &xformOrig); \
	XFORM xformScale = { \
		xformOrig.eM11 * g_Scale, xformOrig.eM12 * g_Scale, \
		xformOrig.eM21 * g_Scale, xformOrig.eM22 * g_Scale, \
		xformOrig.eDx  * g_Scale, xformOrig.eDy  * g_Scale  \
	}; \
	SetWorldTransform(hdc, &xformScale); \
	BOOL b = ORIG_##name vars; \
	SetWorldTransform(hdc, &xformOrig); \
	if(grmode != GM_ADVANCED) \
		SetGraphicsMode(hdc, grmode); \
	return b; \
	}

/*
HOOK_DEFINE(user32.dll, DWORD, GetTabbedTextExtentA, (HDC hdc, LPCSTR lpString, int nCount, int nTabPositions, CONST LPINT lpnTabStopPositions), (hdc, lpString, nCount, nTabPositions, lpnTabStopPositions))
HOOK_DEFINE(user32.dll, DWORD, GetTabbedTextExtentW, (HDC hdc, LPCWSTR lpString, int nCount, int nTabPositions, CONST LPINT lpnTabStopPositions), (hdc, lpString, nCount, nTabPositions, lpnTabStopPositions))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentExPointA, (HDC hdc, LPCSTR lpszStr, int cchString, int nMaxExtent, LPINT lpnFit, LPINT lpDx, LPSIZE lpSize), (hdc, lpszStr, cchString, nMaxExtent, lpnFit, lpDx, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentExPointW, (HDC hdc, LPCWSTR lpszStr, int cchString, int nMaxExtent, LPINT lpnFit, LPINT lpDx, LPSIZE lpSize), (hdc, lpszStr, cchString, nMaxExtent, lpnFit, lpDx, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentExPointI, (HDC hdc, LPWORD pgiIn, int cgi, int nMaxExtent, LPINT lpnFit, LPINT lpDx, LPSIZE lpSize), (hdc, pgiIn, cgi, nMaxExtent, lpnFit, lpDx, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentPointA, (HDC hdc, LPCSTR lpString, int cbString, LPSIZE lpSize), (hdc, lpString, cbString, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentPointW, (HDC hdc, LPCWSTR lpString, int cbString, LPSIZE lpSize), (hdc, lpString, cbString, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentPointI, (HDC hdc, LPWORD pgiIn, int cgi, LPSIZE lpSize), (hdc, pgiIn, cgi, lpSize))
*/

//firefox
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentPoint32A, (HDC hdc, LPCSTR lpString, int cbString, LPSIZE lpSize), (hdc, lpString, cbString, lpSize))
HOOK_DEFINE(gdi32.dll, BOOL, GetTextExtentPoint32W, (HDC hdc, LPCWSTR lpString, int cbString, LPSIZE lpSize), (hdc, lpString, cbString, lpSize))
#undef HOOK_DEFINE


// ȊȂ RpC̍œK{CPU̐ǂݎsŉƂȂ񂩂̂H
// LookupTablex3͂Ƃ肠~߁BŃx`Ă݂B
// (2006/09/22) TusNZ_Oɖ߂Ă݂
//              EBhEd˂ƂȂǂɏoF͂炸...
void  ScaleDIB(BYTE* lpPixels, int width, int height)
{
	int  srcWidthBytes = g_Cache.dibSize.cx * 4;
	int  srcSkipBytes  = srcWidthBytes - (width*4);
	int  dstSkipBytes  = srcWidthBytes - (width*4 / g_Scale);
#if defined(__SSE2__) || defined(__SSE3__)
	__m128i PZERO = _mm_setzero_si128();
#endif
	register BYTE* dst  = lpPixels;
	register BYTE* src0 = lpPixels;
	width  /= g_Scale;
	height /= g_Scale;

	if(g_Scale == 4) {
#if defined(__SSE3__)
		for(int y=0 ; y < height ; y++) {
			register BYTE* src1 = src0 + srcWidthBytes;
			register BYTE* src2 = src1 + srcWidthBytes;
			register BYTE* src3 = src2 + srcWidthBytes;
			int x;
			int widtha;
			widtha = width&(~3);
			for(x=0 ; x < widtha ; x+=4) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6;
				XMM0 = _mm_lddqu_si128((__m128i*)(src0   ));
				XMM1 = _mm_lddqu_si128((__m128i*)(src1   ));
				XMM4 = _mm_lddqu_si128((__m128i*)(src2   ));
				XMM5 = _mm_lddqu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM6 = _mm_lddqu_si128((__m128i*)(src0+16));
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM5 = _mm_lddqu_si128((__m128i*)(src1+16));
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM3 = _mm_lddqu_si128((__m128i*)(src2+16));
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM4 = _mm_lddqu_si128((__m128i*)(src3+16));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = XMM6;
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM5;
				XMM6 = _mm_unpacklo_epi8(XMM6, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_add_epi16(XMM6, XMM5);
				XMM5 = XMM3;
				XMM1 = _mm_add_epi16(XMM1, XMM2);
				XMM2 = XMM4;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM4);
				XMM4 = _mm_lddqu_si128((__m128i*)(src0+32));
				XMM5 = _mm_add_epi16(XMM5, XMM2);
				XMM2 = _mm_lddqu_si128((__m128i*)(src1+32));
				XMM6 = _mm_add_epi16(XMM6, XMM3);
				XMM3 = _mm_lddqu_si128((__m128i*)(src2+32));
				XMM1 = _mm_add_epi16(XMM1, XMM5);
				XMM5 = _mm_lddqu_si128((__m128i*)(src3+32));
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM0;
				XMM0 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM0), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(3,2,1,0)));
				XMM1 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM1), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(1,0,3,2)));
				XMM6 = XMM4;
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM2;
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM2 = _mm_unpacklo_epi8(XMM2, PZERO);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				_mm_storel_epi64((__m128i*)(dst   ), XMM0);
				XMM0 = _mm_lddqu_si128((__m128i*)(src0+48));
				XMM4 = _mm_add_epi16(XMM4, XMM2);
				XMM2 = XMM3;
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM5;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM5);
				XMM5 = _mm_lddqu_si128((__m128i*)(src1+48));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = _mm_lddqu_si128((__m128i*)(src2+48));
				XMM4 = _mm_add_epi16(XMM4, XMM3);
				XMM3 = _mm_lddqu_si128((__m128i*)(src3+48));
				XMM6 = _mm_add_epi16(XMM6, XMM2);
				XMM2 = XMM0;
				XMM4 = _mm_add_epi16(XMM4, XMM6);
				XMM6 = XMM5;
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM5);
				XMM5 = XMM1;
				XMM2 = _mm_add_epi16(XMM2, XMM6);
				XMM6 = XMM3;
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM5 = _mm_add_epi16(XMM5, XMM6);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM2 = _mm_add_epi16(XMM2, XMM5);
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM4;
				XMM4 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM4), _mm_castsi128_ps(XMM0), _MM_SHUFFLE(3,2,1,0)));
				XMM2 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM2), _mm_castsi128_ps(XMM0), _MM_SHUFFLE(1,0,3,2)));
				XMM4 = _mm_add_epi16(XMM4, XMM2);
				XMM4 = _mm_srli_epi16(XMM4, 4);
				XMM4 = _mm_packus_epi16(XMM4, XMM4);
				_mm_storel_epi64((__m128i*)(dst+ 8), XMM4);
				dst+=16; src0+=4*16; src1+=4*16; src2+=4*16; src3+=4*16;
			}
			widtha = width&(~1);
			for(    ; x < widtha ; x+=2) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6;
				XMM0 = _mm_lddqu_si128((__m128i*)(src0   ));
				XMM1 = _mm_lddqu_si128((__m128i*)(src1   ));
				XMM4 = _mm_lddqu_si128((__m128i*)(src2   ));
				XMM5 = _mm_lddqu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM6 = _mm_lddqu_si128((__m128i*)(src0+16));
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM5 = _mm_lddqu_si128((__m128i*)(src1+16));
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM3 = _mm_lddqu_si128((__m128i*)(src2+16));
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM4 = _mm_lddqu_si128((__m128i*)(src3+16));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = XMM6;
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM5;
				XMM6 = _mm_unpacklo_epi8(XMM6, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_add_epi16(XMM6, XMM5);
				XMM5 = XMM3;
				XMM1 = _mm_add_epi16(XMM1, XMM2);
				XMM2 = XMM4;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM4);
				XMM5 = _mm_add_epi16(XMM5, XMM2);
				XMM6 = _mm_add_epi16(XMM6, XMM3);
				XMM1 = _mm_add_epi16(XMM1, XMM5);
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM0;
				XMM0 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM0), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(3,2,1,0)));
				XMM1 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM1), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(1,0,3,2)));
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				_mm_storel_epi64((__m128i*)(dst   ), XMM0);
				dst+=8; src0+=4*8; src1+=4*8; src2+=4*8; src3+=4*8;
			}
			for(    ; x < width ; x++) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5;
				XMM0 = _mm_lddqu_si128((__m128i*)(src0   ));
				XMM1 = _mm_lddqu_si128((__m128i*)(src1   ));
				XMM4 = _mm_lddqu_si128((__m128i*)(src2   ));
				XMM5 = _mm_lddqu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = _mm_castps_si128(_mm_movehl_ps(_mm_castsi128_ps(XMM2), _mm_castsi128_ps(XMM0)));
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				_mm_store_ss((float*)(dst   ), _mm_castsi128_ps(XMM0));
				dst+=4; src0+=4*4; src1+=4*4; src2+=4*4; src3+=4*4;
			}
			dst += dstSkipBytes;
			src0 = src3 + srcSkipBytes;
		}
#elif defined(__SSE2__)
		for(int y=0 ; y < height ; y++) {
			register BYTE* src1 = src0 + srcWidthBytes;
			register BYTE* src2 = src1 + srcWidthBytes;
			register BYTE* src3 = src2 + srcWidthBytes;
			int x;
			int widtha;
			widtha = width&(~3);
			for(x=0 ; x < widtha ; x+=4) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6;
				XMM0 = _mm_loadu_si128((__m128i*)(src0   ));
				XMM1 = _mm_loadu_si128((__m128i*)(src1   ));
				XMM4 = _mm_loadu_si128((__m128i*)(src2   ));
				XMM5 = _mm_loadu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM6 = _mm_loadu_si128((__m128i*)(src0+16));
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM5 = _mm_loadu_si128((__m128i*)(src1+16));
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM3 = _mm_loadu_si128((__m128i*)(src2+16));
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM4 = _mm_loadu_si128((__m128i*)(src3+16));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = XMM6;
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM5;
				XMM6 = _mm_unpacklo_epi8(XMM6, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_add_epi16(XMM6, XMM5);
				XMM5 = XMM3;
				XMM1 = _mm_add_epi16(XMM1, XMM2);
				XMM2 = XMM4;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM4);
				XMM4 = _mm_loadu_si128((__m128i*)(src0+32));
				XMM5 = _mm_add_epi16(XMM5, XMM2);
				XMM2 = _mm_loadu_si128((__m128i*)(src1+32));
				XMM6 = _mm_add_epi16(XMM6, XMM3);
				XMM3 = _mm_loadu_si128((__m128i*)(src2+32));
				XMM1 = _mm_add_epi16(XMM1, XMM5);
				XMM5 = _mm_loadu_si128((__m128i*)(src3+32));
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM0;
				XMM0 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM0), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(3,2,1,0)));
				XMM1 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM1), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(1,0,3,2)));
				XMM6 = XMM4;
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM2;
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM2 = _mm_unpacklo_epi8(XMM2, PZERO);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				_mm_storel_epi64((__m128i*)(dst   ), XMM0);
				XMM0 = _mm_loadu_si128((__m128i*)(src0+48));
				XMM4 = _mm_add_epi16(XMM4, XMM2);
				XMM2 = XMM3;
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM5;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM5);
				XMM5 = _mm_loadu_si128((__m128i*)(src1+48));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = _mm_loadu_si128((__m128i*)(src2+48));
				XMM4 = _mm_add_epi16(XMM4, XMM3);
				XMM3 = _mm_loadu_si128((__m128i*)(src3+48));
				XMM6 = _mm_add_epi16(XMM6, XMM2);
				XMM2 = XMM0;
				XMM4 = _mm_add_epi16(XMM4, XMM6);
				XMM6 = XMM5;
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM5);
				XMM5 = XMM1;
				XMM2 = _mm_add_epi16(XMM2, XMM6);
				XMM6 = XMM3;
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM6 = _mm_unpackhi_epi8(XMM6, PZERO);
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM5 = _mm_add_epi16(XMM5, XMM6);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM2 = _mm_add_epi16(XMM2, XMM5);
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM4;
				XMM4 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM4), _mm_castsi128_ps(XMM0), _MM_SHUFFLE(3,2,1,0)));
				XMM2 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM2), _mm_castsi128_ps(XMM0), _MM_SHUFFLE(1,0,3,2)));
				XMM4 = _mm_add_epi16(XMM4, XMM2);
				XMM4 = _mm_srli_epi16(XMM4, 4);
				XMM4 = _mm_packus_epi16(XMM4, XMM4);
				_mm_storel_epi64((__m128i*)(dst+ 8), XMM4);
				dst+=16; src0+=4*16; src1+=4*16; src2+=4*16; src3+=4*16;
			}
			widtha = width&(~1);
			for(    ; x < widtha ; x+=2) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6;
				XMM0 = _mm_loadu_si128((__m128i*)(src0   ));
				XMM1 = _mm_loadu_si128((__m128i*)(src1   ));
				XMM4 = _mm_loadu_si128((__m128i*)(src2   ));
				XMM5 = _mm_loadu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM6 = _mm_loadu_si128((__m128i*)(src0+16));
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM5 = _mm_loadu_si128((__m128i*)(src1+16));
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM3 = _mm_loadu_si128((__m128i*)(src2+16));
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM4 = _mm_loadu_si128((__m128i*)(src3+16));
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM1 = XMM6;
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = XMM5;
				XMM6 = _mm_unpacklo_epi8(XMM6, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM6 = _mm_add_epi16(XMM6, XMM5);
				XMM5 = XMM3;
				XMM1 = _mm_add_epi16(XMM1, XMM2);
				XMM2 = XMM4;
				XMM3 = _mm_unpacklo_epi8(XMM3, PZERO);
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpackhi_epi8(XMM5, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_add_epi16(XMM3, XMM4);
				XMM5 = _mm_add_epi16(XMM5, XMM2);
				XMM6 = _mm_add_epi16(XMM6, XMM3);
				XMM1 = _mm_add_epi16(XMM1, XMM5);
				XMM6 = _mm_add_epi16(XMM6, XMM1);
				XMM1 = XMM0;
				XMM0 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM0), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(3,2,1,0)));
				XMM1 = _mm_castps_si128(_mm_shuffle_ps(
					_mm_castsi128_ps(XMM1), _mm_castsi128_ps(XMM6), _MM_SHUFFLE(1,0,3,2)));
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				_mm_storel_epi64((__m128i*)(dst   ), XMM0);
				dst+=8; src0+=4*8; src1+=4*8; src2+=4*8; src3+=4*8;
			}
			for(    ; x < width ; x++) {
				__m128i XMM0, XMM1, XMM2, XMM3, XMM4, XMM5;
				XMM0 = _mm_loadu_si128((__m128i*)(src0   ));
				XMM1 = _mm_loadu_si128((__m128i*)(src1   ));
				XMM4 = _mm_loadu_si128((__m128i*)(src2   ));
				XMM5 = _mm_loadu_si128((__m128i*)(src3   ));
				XMM2 = XMM0;
				XMM3 = XMM1;
				XMM0 = _mm_unpacklo_epi8(XMM0, PZERO);
				XMM1 = _mm_unpacklo_epi8(XMM1, PZERO);
				XMM2 = _mm_unpackhi_epi8(XMM2, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM0 = _mm_add_epi16(XMM0, XMM1);
				XMM1 = XMM4;
				XMM2 = _mm_add_epi16(XMM2, XMM3);
				XMM3 = XMM5;
				XMM4 = _mm_unpacklo_epi8(XMM4, PZERO);
				XMM5 = _mm_unpacklo_epi8(XMM5, PZERO);
				XMM1 = _mm_unpackhi_epi8(XMM1, PZERO);
				XMM3 = _mm_unpackhi_epi8(XMM3, PZERO);
				XMM4 = _mm_add_epi16(XMM4, XMM5);
				XMM1 = _mm_add_epi16(XMM1, XMM3);
				XMM0 = _mm_add_epi16(XMM0, XMM4);
				XMM2 = _mm_add_epi16(XMM2, XMM1);
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM2 = _mm_castps_si128(_mm_movehl_ps(_mm_castsi128_ps(XMM2), _mm_castsi128_ps(XMM0)));
				XMM0 = _mm_add_epi16(XMM0, XMM2);
				XMM0 = _mm_srli_epi16(XMM0, 4);
				XMM0 = _mm_packus_epi16(XMM0, XMM0);
				_mm_store_ss((float*)(dst   ), _mm_castsi128_ps(XMM0));
				dst+=4; src0+=4*4; src1+=4*4; src2+=4*4; src3+=4*4;
			}
			dst += dstSkipBytes;
			src0 = src3 + srcSkipBytes;
		}
#else

		for(int y=0 ; y < height ; y++) {
			register BYTE* src1 = src0 + srcWidthBytes;
			register BYTE* src2 = src1 + srcWidthBytes;
			register BYTE* src3 = src2 + srcWidthBytes;
			for(int x=0 ; x < width ; x++) {
				dst[0]=(src0[0] + src0[4] + src0[8] + src0[12] +
					src1[0] + src1[4] + src1[8] + src1[12] +
					src2[0] + src2[4] + src2[8] + src2[12] +
					src3[0] + src3[4] + src3[8] + src3[12]) >> 4;
				dst[1]=(src0[1] + src0[5] + src0[9] + src0[13] +
					src1[1] + src1[5] + src1[9] + src1[13] +
					src2[1] + src2[5] + src2[9] + src2[13] +
					src3[1] + src3[5] + src3[9] + src3[13]) >> 4;
				dst[2]=(src0[2] + src0[6] + src0[10] + src0[14] +
					src1[2] + src1[6] + src1[10] + src1[14] +
					src2[2] + src2[6] + src2[10] + src2[14] +
					src3[2] + src3[6] + src3[10] + src3[14]) >> 4;
				dst+=4; src0+=4*4; src1+=4*4; src2+=4*4; src3+=4*4;
			}
			dst += dstSkipBytes;
			src0 = src3 + srcSkipBytes;
		}
#endif
	}
	else if(g_Scale == 3) {
		if (g_UseSubPixel==0) {
			for(int y=0 ; y < height ; y++) {
				// Don't use subpixels
				register BYTE* src1 = src0 + srcWidthBytes;
				register BYTE* src2 = src1 + srcWidthBytes;
				for(int x=0 ; x < width; x++) {
					dst[0]=(src0[0] + src0[4] + src0[8] +
						src1[0] + src1[4] + src1[8] +
						src2[0] + src2[4] + src2[8]) / 9;
					dst[1]=(src0[1] + src0[5] + src0[9] +
						src1[1] + src1[5] + src1[9] +
						src2[1] + src2[5] + src2[9]) / 9;
					dst[2]=(src0[2] + src0[6] + src0[10] +
						src1[2] + src1[6] + src1[10] +
						src2[2] + src2[6] + src2[10]) / 9;
					dst+=4; src0+=4*3; src1+=4*3; src2+=4*3;
				}
				dst += dstSkipBytes;
				src0 = src2 + srcSkipBytes;
			}
		} else {
			if (g_SubPixelDirection==0) {
				for(int y=0 ; y < height ; y++) {
					// RGB
					register BYTE* src1 = src0 + srcWidthBytes;
					register BYTE* src2 = src1 + srcWidthBytes;					
					dst[0]=(src0[0] + src0[4] + src0[8] +
						src1[0] + src1[4] + src1[8] +
						src2[0] + src2[4] + src2[8]) / 9;
					dst[1]=(src0[1] + src0[5] + src0[9] +
						src1[1] + src1[5] + src1[9] +
						src2[1] + src2[5] + src2[9]) / 9;
					dst[2]=(src0[2] + src0[6] + src0[10] +
						src1[2] + src1[6] + src1[10] +
						src2[2] + src2[6] + src2[10]) / 9;
					dst+=4;src0+=4*3; src1+=4*3; src2+=4*3;
					for(int x=1 ; x < width-1; x++) {
						dst[0]=(src0[8] + src0[12] + src0[16] +
							src1[8] + src1[12] + src1[16] +
							src2[8] + src2[12] + src2[16]) / 9;
						dst[1]=(src0[5] + src0[9] + src0[13] +
							src1[5] + src1[9] + src1[13] +
							src2[5] + src2[9] + src2[13]) / 9;
						dst[2]=(src0[2] + src0[6] + src0[10] +
							src1[2] + src1[6] + src1[10] +
							src2[2] + src2[6] + src2[10]) / 9;
						dst+=4; src0+=4*3; src1+=4*3; src2+=4*3;
					}
					dst[0]=(src0[0] + src0[4] + src0[8] +
						src1[0] + src1[4] + src1[8] +
						src2[0] + src2[4] + src2[8]) / 9;
					dst[1]=(src0[1] + src0[5] + src0[9] +
						src1[1] + src1[5] + src1[9] +
						src2[1] + src2[5] + src2[9]) / 9;
					dst[2]=(src0[2] + src0[6] + src0[10] +
						src1[2] + src1[6] + src1[10] +
						src2[2] + src2[6] + src2[10]) / 9;
					//dst+=4; src0+=4*3; src1+=4*3; src2+=4*3;
					dst += dstSkipBytes + 4;
					src0 = src2 + (4*3) + srcSkipBytes;
				}
			} else {
				for(int y=0 ; y < height ; y++) {
					// BGR
					register BYTE* src1 = src0 + srcWidthBytes;
					register BYTE* src2 = src1 + srcWidthBytes;					
					dst[0]=(src0[0] + src0[4] + src0[8] +
						src1[0] + src1[4] + src1[8] +
						src2[0] + src2[4] + src2[8]) / 9;
					dst[1]=(src0[1] + src0[5] + src0[9] +
						src1[1] + src1[5] + src1[9] +
						src2[1] + src2[5] + src2[9]) / 9;
					dst[2]=(src0[2] + src0[6] + src0[10] +
						src1[2] + src1[6] + src1[10] +
						src2[2] + src2[6] + src2[10]) / 9;
					dst+=4;src0+=4*3; src1+=4*3; src2+=4*3;
					for(int x=1 ; x < width-1; x++) {
						dst[0]=(src0[0] + src0[4] + src0[8] +
							src1[0] + src1[4] + src1[8] +
							src2[0] + src2[4] + src2[8]) / 9;
						dst[1]=(src0[5] + src0[9] + src0[13] +
							src1[5] + src1[9] + src1[13] +
							src2[5] + src2[9] + src2[13]) / 9;
						dst[2]=(src0[10] + src0[14] + src0[18] +
							src1[10] + src1[14] + src1[18] +
							src2[10] + src2[14] + src2[18]) / 9;
						dst+=4; src0+=4*3; src1+=4*3; src2+=4*3;
					}
					dst[0]=(src0[0] + src0[4] + src0[8] +
						src1[0] + src1[4] + src1[8] +
						src2[0] + src2[4] + src2[8]) / 9;
					dst[1]=(src0[1] + src0[5] + src0[9] +
						src1[1] + src1[5] + src1[9] +
						src2[1] + src2[5] + src2[9]) / 9;
					dst[2]=(src0[2] + src0[6] + src0[10] +
						src1[2] + src1[6] + src1[10] +
						src2[2] + src2[6] + src2[10]) / 9;
					//dst+=4; src0+=4*3; src1+=4*3; src2+=4*3;
					dst += dstSkipBytes + 4;
					src0 = src2 + (4*3) + srcSkipBytes;
				}
			}
		}
	}
	else if(g_Scale == 2) {
		for(int y=0 ; y < height ; y++) {
			register BYTE* src1 = src0 + srcWidthBytes;
			for(int x=0 ; x < width ; x++) {
				dst[0]=(src0[0] + src0[4] +
					src1[0] + src1[4]) >> 2;
				dst[1]=(src0[1] + src0[5] +
					src1[1] + src1[5]) >> 2;
				dst[2]=(src0[2] + src0[6] +
					src1[2] + src1[6]) >> 2;
				dst+=4; src0+=4*2; src1+=4*2;
			}
			dst += dstSkipBytes;
			src0 = src1 + srcSkipBytes;
		}
	}
}


HDC  CreateCacheDC()
{
	if(!g_Cache.hdc)
		g_Cache.hdc = CreateCompatibleDC(NULL);
	return g_Cache.hdc;
}

HBITMAP  CreateCacheDIB(HDC hdc, int width, int height, BYTE** lpPixels)
{
	if(g_Cache.dibSize.cx < width || g_Cache.dibSize.cy < height) {
		if(g_Cache.hbmp)
			DeleteObject(g_Cache.hbmp);

		if(g_Cache.dibSize.cx < width)
			g_Cache.dibSize.cx = width;
		if(g_Cache.dibSize.cy < height)
			g_Cache.dibSize.cy = height;

		BITMAPINFOHEADER bmiHeader;
		bmiHeader.biSize = sizeof(bmiHeader);
		bmiHeader.biWidth = g_Cache.dibSize.cx;
		bmiHeader.biHeight = -g_Cache.dibSize.cy;
		bmiHeader.biPlanes = 1;
		bmiHeader.biBitCount = 32;
		bmiHeader.biCompression = BI_RGB;
		bmiHeader.biSizeImage = 0;
		bmiHeader.biXPelsPerMeter = 0;
		bmiHeader.biYPelsPerMeter = 0;
		bmiHeader.biClrUsed = 0;
		bmiHeader.biClrImportant = 0;
		g_Cache.hbmp = CreateDIBSection(hdc, (CONST BITMAPINFO*)&bmiHeader, DIB_RGB_COLORS, (LPVOID*)&g_Cache.lpPixels, NULL, 0);
	}
	*lpPixels = g_Cache.lpPixels;
	return g_Cache.hbmp;
}

int div_ceil(int a, int b)
{
	if(a % b)
		return (a>0)? (a/b+1): (a/b-1);
	return a / b;
}

// Windows 2000p
BOOL WINAPI IMPL_TextOutA(HDC hdc, int nXStart, int nYStart, LPCSTR lpString, UINT cbString)
{
	if (g_IsWinXPorLater) {
		return ORIG_TextOutA(hdc, nXStart, nYStart, lpString, cbString);
	} else {
		return IMPL_ExtTextOutA(hdc, nXStart, nYStart, NULL, NULL, lpString, cbString, NULL);
	}
}

BOOL WINAPI IMPL_TextOutW(HDC hdc, int nXStart, int nYStart, LPCWSTR lpString, UINT cbString)
{
	if (g_IsWinXPorLater) {
		return ORIG_TextOutW(hdc, nXStart, nYStart, lpString, cbString);
	} else {
		return IMPL_ExtTextOutW(hdc, nXStart, nYStart, NULL, NULL, lpString, cbString, NULL);
	}
}

void ConvertDxArray(UINT CodePage, LPCSTR lpString, const int* lpDxA, int cbString, int* lpDxW, int)
{
	LPCSTR lpEnd = lpString + cbString;
	while(lpString < lpEnd) {
		if(IsDBCSLeadByteEx(CodePage, *lpString)) {
			*lpDxW = *lpDxA++;
			*lpDxW += *lpDxA++;
			lpDxW++;
			lpString++;
			if(lpString == lpEnd)
				break;
		}
		else
			*lpDxW++ = *lpDxA++;
		lpString++;
	}
}

// ANSI->UnicodeɕϊExtTextOutWɓExtTextOutAiWindows 2000pj
BOOL WINAPI IMPL_ExtTextOutA(HDC hdc, int nXStart, int nYStart, UINT fuOptions, CONST RECT *lprc, LPCSTR lpString, UINT cbString, CONST INT *lpDx)
{
	if (g_IsWinXPorLater) {
		return ORIG_ExtTextOutA(hdc, nXStart, nYStart, fuOptions, lprc, lpString, cbString, lpDx);
	} else {
		LPWSTR lpszUnicode;
		int bufferLength;
		BOOL result;
		bufferLength = MultiByteToWideChar(CP_ACP, 0, lpString, cbString, 0, 0);
	//	lpszUnicode = (LPWSTR)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, bufferLength * 2);
		lpszUnicode = (LPWSTR)malloc(bufferLength * 2);
		MultiByteToWideChar(CP_ACP, 0, lpString, cbString, lpszUnicode, bufferLength);

		UINT cp = _getmbcp();
		int* lpDxW = NULL;
		if(lpDx && cbString && cp) {
			lpDxW = (int*)calloc(sizeof(int), cbString);
			ConvertDxArray(cp, lpString, lpDx, cbString, lpDxW, 0);
			lpDx = lpDxW;
		}

		result = IMPL_ExtTextOutW(hdc, nXStart, nYStart, fuOptions, lprc, (LPCWSTR)lpszUnicode, bufferLength, lpDx);
	//	HeapFree(GetProcessHeap(), HEAP_NO_SERIALIZE, lpszUnicode);
		free(lpszUnicode);
		free(lpDxW);	// free(NULL)͖薳ۏ؂Ă͂
		return result;
	}
}

// LCExtTextOutiДj̎
BOOL WINAPI IMPL_ExtTextOutW(HDC hdc, int nXStart, int nYStart, UINT fuOptions, CONST RECT *lprc, LPCWSTR lpString, UINT cbString, CONST INT *lpDx)
{
	CacheLock  __lock;
	int   error = 0;

	POINT curPos = { nXStart, nYStart };
	POINT destPos;
	SIZE  destSize;
	POINT canvasPos;
	SIZE  canvasSize;

	HDC     hCanvasDC = NULL;
	HGDIOBJ hPrevFont = NULL;
	HBITMAP hBmp = NULL;
	BYTE*   lpPixels = NULL;

	XFORM  xformNoTrns;
	XFORM  xformScale = {1.0f,0.0f,  0.0f,1.0f,  0.0f,0.0f};
	UINT   align;
	SIZE   textSize;

	WCHAR  szFaceName[LF_FACESIZE] = L"";
	
	if(lpString == NULL || cbString == 0 || g_Quality==0) {
		error = 11;//throw
	}
	if(!error) {
		GetTextFace(hdc,LF_FACESIZE,szFaceName);
		if(IsFontExcluded(szFaceName)) {
			error = 14;	// the font has been excluded
		}
	}
	if(!error) {
		hCanvasDC = CreateCacheDC();
		if(!hCanvasDC)
			error = 1;
	}
	// get cursor / xform
	if(!error) {
		align = GetTextAlign(hdc);
		if(align & TA_UPDATECP) {
			GetCurrentPositionEx(hdc, &curPos);
		}

		GetWorldTransform(hdc, &xformNoTrns);
		if(xformNoTrns.eM12 < -0.1f || 0.1f < xformNoTrns.eM12 ||
		   xformNoTrns.eM21 < -0.1f || 0.1f < xformNoTrns.eM21) {
			error = 12;// throw rotation
		}
		xformNoTrns.eDx = 0.0f; //clear translation
		xformNoTrns.eDy = 0.0f;
		xformScale.eM11 = xformNoTrns.eM11 * g_Scale;//scaling
		xformScale.eM22 = xformNoTrns.eM22 * g_Scale;
		SetGraphicsMode(hCanvasDC, GM_ADVANCED);
		SetWorldTransform(hCanvasDC, &xformScale);
	}
	//copy font
	if(!error) {
		// ̕H
		hPrevFont = SelectObject(hCanvasDC, GetCurrentObject(hdc, OBJ_FONT));
/*		HGDIOBJ src = SelectObject(hdc, GetStockObject(SYSTEM_FONT));
		if(!src)
			error = 2;
		else {
			SelectObject(hdc, src);
			hPrevFont = SelectObject(hCanvasDC, src);
		}*/
	}

	//GetTextExtentPointnSetTextCharacterExtraKvƂ炵
	SetTextCharacterExtra(hCanvasDC, GetTextCharacterExtra(hdc));
	//text size
	if(!error) {
		if(fuOptions & ETO_GLYPH_INDEX)
			//ORIG_GetTextExtentPointI(hCanvasDC, (LPWORD)lpString, (int)cbString, &textSize);
			GetTextExtentPointI(hCanvasDC, (LPWORD)lpString, (int)cbString, &textSize);
		else
			//ORIG_GetTextExtentPointW(hCanvasDC, lpString, (int)cbString, &textSize);
//			GetTextExtentPointW(hCanvasDC, lpString, (int)cbString, &textSize);

			//GetTextExtentPointWTeraPad on win2kŐ`悳Ȃ
			//rbg}bvtHgTransformɑΉĂȂH
			GetTextExtentExPointW(hCanvasDC, lpString, (int)cbString, 0, NULL, NULL, &textSize);
		if(lpDx) {
			textSize.cx = 0;
			if(fuOptions & ETO_PDY)
				for(UINT i=0; i < cbString; textSize.cx += lpDx[i],i+=2);
			else
				for(UINT i=0; i < cbString; textSize.cx += lpDx[i++]);
		}

		// C^bÑXg𑫂(\̖̂OABC
		// Ql: http://bugzilla.mozilla.gr.jp/show_bug.cgi?id=3253
		TEXTMETRIC tm;
		GetTextMetrics(hdc, &tm);		
		if (tm.tmItalic) {
			ABC abcWidth={0, 0, 0};
			GetCharABCWidthsW(hdc, lpString[cbString-1], lpString[cbString-1], &abcWidth);
			textSize.cx += tm.tmOverhang - abcWidth.abcC;
		}

		//if(textSize.cy * xformNoTrns.eM22 >= 24.0f) //max 24pixel
		if (g_MaxHeight > 0) {
			if(textSize.cy * xformNoTrns.eM22 >= g_MaxHeight) //max 26pixel
				error = 13;//throw large size
		}
	}
	//rectangle
	if(!error) {
		//trace(L"OrigTextSize=%d %d\n", textSize.cx, textSize.cy);
		//trace(L"OrigCursor=%d %d\n", curPos.x, curPos.y);

		RECT rc;
		UINT horiz = align & (TA_LEFT|TA_RIGHT|TA_CENTER);
		UINT vert  = align & (TA_BASELINE|TA_TOP);//TA_BOTTOM
		if(horiz == TA_CENTER) {
			rc.left  = curPos.x - div_ceil(textSize.cx, 2);
			rc.right = curPos.x + div_ceil(textSize.cx, 2);
			//no move
		}
		else if(horiz == TA_RIGHT) {
			rc.left  = curPos.x - textSize.cx;
			rc.right = curPos.x;
			curPos.x -= textSize.cx;//move pos
		}
		else {
			rc.left  = curPos.x;
			rc.right = curPos.x + textSize.cx;
			curPos.x += textSize.cx;//move pos
		}
		if(vert == TA_BASELINE) {
			TEXTMETRIC metric = {0};
			GetTextMetrics(hCanvasDC, &metric);
			rc.top = curPos.y - metric.tmAscent;
			rc.bottom = curPos.y + metric.tmDescent;
			//trace(L"ascent=%d descent=%d\n", metric.tmAscent, metric.tmDescent);
		}
		else {
			rc.top = curPos.y;
			rc.bottom = curPos.y + textSize.cy;
		}

		canvasPos.x = canvasPos.y = 0;

		//clipping
		if(lprc && (fuOptions & ETO_CLIPPED)) {
			if(rc.left < lprc->left) {
				canvasPos.x -= div_ceil( (int)((lprc->left - rc.left) * xformScale.eM11), g_Scale);
				rc.left = lprc->left;
			}
			if(rc.right > lprc->right) {
				if(horiz == TA_RIGHT)
					canvasPos.x -= div_ceil( (int)((rc.right - lprc->right) * xformScale.eM11), g_Scale);
				rc.right = lprc->right;
			}
			if(rc.top < lprc->top) {
				canvasPos.y -= div_ceil( (int)((lprc->top - rc.top) * xformScale.eM22), g_Scale);
				rc.top = lprc->top;
			}
			if(rc.bottom > lprc->bottom) {
				rc.bottom = lprc->bottom;
			}
			//trace(L"ClipRect=%d %d %d %d\n", lprc->left, lprc->top, lprc->right, lprc->bottom);
		}

		destPos.x = rc.left;
		destPos.y = rc.top;
		destSize.cx = rc.right - rc.left;
		destSize.cy = rc.bottom - rc.top;
		canvasSize.cx = (int)(destSize.cx * xformScale.eM11);
		canvasSize.cy = (int)(destSize.cy * xformScale.eM22);

		//trace(L"MovedCursor=%d %d\n", curPos.x, curPos.y);
		//trace(L"TargetRect=%d %d %d %d\n", rc.left, rc.top, rc.right, rc.bottom);
		//trace(L"DestPos=%dx%d Size=%dx%d\n", destPos.x, destPos.y, destSize.cx, destSize.cy);
		//trace(L"CanvasPos=%dx%d Size=%dx%d\n", canvasPos.x, canvasPos.y, canvasSize.cx, canvasSize.cy);

		if(destSize.cx < 1 || destSize.cy < 1)
			error = 14; //throw no area
	}
	//bitmap
	if(!error) {
		hBmp = CreateCacheDIB(hCanvasDC, canvasSize.cx, canvasSize.cy, &lpPixels);
		if(!hBmp)
			error = 3;
	}
	if(!error) {
		HGDIOBJ hPrevBmp = SelectObject(hCanvasDC, hBmp);

		BOOL fillrect = (lprc && (fuOptions & ETO_OPAQUE)) ? TRUE : FALSE;

		//clear bitmap
		if(fillrect || GetBkMode(hdc) == OPAQUE) {
			SetBkMode(hCanvasDC, OPAQUE);
			COLORREF  bgcolor = GetBkColor(hdc); //ƂwiF
			SetBkColor(hCanvasDC, bgcolor);

			//Brush荂炵(MS)
//			RECT rc = { 0,0, destSize.cx, destSize.cy };//WorldTransformĂ邩destSizeŗǂ
//			ORIG_ExtTextOutW(hCanvasDC, 0,0, ETO_OPAQUE, &rc, NULL, 0, NULL);
			//DIBڑ (܂ςĂȂEEE)
			SelectObject(hCanvasDC, hPrevBmp);
			RECT rc = { 0, 0, canvasSize.cx, canvasSize.cy };
			CacheFillSolidRect(bgcolor, &rc);
			SelectObject(hCanvasDC, hBmp);

			if(fillrect) {
				ORIG_ExtTextOutW(hdc, 0,0, ETO_OPAQUE, lprc, NULL, 0, NULL);
			}
		}
		else {
			SetBkMode(hCanvasDC, TRANSPARENT);
			BitBlt(hCanvasDC,0,0,destSize.cx,destSize.cy, hdc,destPos.x,destPos.y, SRCCOPY);//WorldTransformĂ邩destSizeŗǂ
		}

		//setup
		SetTextColor(hCanvasDC, GetTextColor(hdc));
		SetTextAlign(hCanvasDC, TA_LEFT | TA_TOP);

		//textout
		for (UINT w=0; w<=g_Weight; w++) ORIG_ExtTextOutW(hCanvasDC, canvasPos.x, canvasPos.y, fuOptions, NULL, lpString, cbString, lpDx);

		//SaveDIB(L"a0.bmp", canvasSize.cx, canvasSize.cy, lpPixels);

		//scale
		ScaleDIB(lpPixels, canvasSize.cx, canvasSize.cy);

		//SaveDIB(L"a1.bmp", canvasSize.cx, canvasSize.cy, lpPixels);

		//blt
		SetWorldTransform(hCanvasDC, &xformNoTrns);
		BitBlt(hdc, destPos.x, destPos.y, destSize.cx, destSize.cy, hCanvasDC, 0,0, SRCCOPY);

		SelectObject(hCanvasDC, hPrevBmp);

		if(align & TA_UPDATECP) {
			MoveToEx(hdc, curPos.x, curPos.y, NULL);
		}
	}

	//if(hBmp) DeleteObject(hBmp);
	if(hPrevFont) SelectObject(hCanvasDC, hPrevFont);
	//if(hCanvasDC) DeleteDC(hCanvasDC);

	if(!error)
		return TRUE;
	return ORIG_ExtTextOutW(hdc, nXStart, nYStart, fuOptions, lprc, lpString, cbString, lpDx);
}

//EOF
